/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.view;

import java.awt.Color;
import java.awt.GridLayout;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import sudoku.controller.SudokuTransport;
import sudoku.model.games.GameType;

/**
 * The panel holding the sudoku.
 * 
 * @author Sebastian Schauer
 */
public class SudokuView extends JPanel {
    
    /**
     * Whether the user has changed the sudoku since the last saving
     */
    private boolean isChanged;
    /**
     * The root size of the sudoku. The amount of rows of a subcell.
     */
    private int root;
    /**
     * The game type of the sudoku.
     */
    private GameType type;
    /**
     * The window holding this sudoku.
     */
    private Window myWindow;
    /**
     * The array of subsquare panels
     */
    SudokuSubSquare[][] subsquares;
    /**
     * The array of cells
     */
    SudokuCell[][] cells;
    
    /**
     * The size of the sudoku. The amount of rows.
     * @return The size
     */
    public int getSudokuSize() {
        return root * root;
    }
    
    /**
     * The type of the displayed sudoku
     * @return 
     */
    public GameType getSudokuType() {
        return type;
    }
    
    /**
     * Set the type of the displayed sudoku
     * @param t the type to set (see model.games.GameType)
     */
    public void setType(String t) {
        switch (t) {
            case "Regular":
                type = GameType.Regular;
                break;
            case "X": type = GameType.X; break;
            case "Color": type = GameType.Color; break;
            case "Hyper": type = GameType.Hyper; break;
        }
        drawBackgrounds(type);
    }
    
    /**
     * Get the value of the cell at (x, y), 0-based.
     * @param x The x-coordinate
     * @param y The y-coordinate
     * @return The value of the cell
     */
    public int getValue(int x, int y) {
        return cells[x][y].getValue();
    }
    
    /**
     * @return Whether the user has changed the sudoku
     */
    public boolean getChanged() {
        return isChanged;
    }

    /**
     * Set the isChanged property of this sudoku
     * @param isChanged Whether the sudoku has changed
     */
    public void setChanged(boolean isChanged) {
        this.isChanged = isChanged;
    }
     
    /**
     * The constructor
     * @param st The sudoku transport object containing the sudoku to display
     * @param win The window displaying the sudoku
     */
    public SudokuView(SudokuTransport st, Window win) {
        myWindow = win;
        root = (int)Math.sqrt(st.size);
        type = st.type;
        isChanged = false;
                
        setLayout(new GridLayout(root, root, 2, 2));
        subsquares = new SudokuSubSquare[root][root];
        cells = new SudokuCell[root * root][root * root];
        
        for (int subx = 0; subx < root; subx++) {
            for (int suby = 0; suby < root; suby++) {
                subsquares[subx][suby] = new SudokuSubSquare(root);
                for (int cellx = 0; cellx < root; cellx++) {
                    for (int celly = 0; celly < root; celly++) {
                        int totalx = subx * root + cellx;
                        int totaly = suby * root + celly;
                        cells[totalx][totaly] = new SudokuCell(root * root, this, totalx, totaly);
                        subsquares[subx][suby].add(cells[totalx][totaly]).setLocation(cellx, celly);
                    }
                }
                this.add(subsquares[subx][suby]).setLocation(subx, suby);
            }
        }
        
        showSudokuValues(st);
        
        setType(st.type.name());
        
        myWindow.bgrType.setSelected(myWindow.btnType.get(type).getModel(), true);
        cells[0][0].requestFocusInWindow();       
        
        setVisible(true);
    }
    
    /**
     * Get the sudoku transport object of the displayed sudoku
     * @return 
     */
    public SudokuTransport getSudokuTransport() {
        int size = root * root;
        SudokuTransport st = new SudokuTransport(root * root);
        st.type = myWindow.getSudokuType();
        for (int i = 0; i < size; i++)
            for (int j = 0; j < size; j++)
                st.values[i][j] = cells[i][j].getValue();
        return st;
    }
    
    /**
     * Whether sudoku may be closed. Returns true if sudoku is unchanged or 
     * if the user wants to discard it.
     * @return Whether sudoku may be closed
     */
    public boolean mayClose() {
        if (isChanged) {
            int r = JOptionPane.showConfirmDialog(null, "This sudoku is unsaved. Save changes?", "Sudoku unsaved", JOptionPane.YES_NO_OPTION);
            return (r == JOptionPane.NO_OPTION);
        }
        return true;
    }
    
    /**
     * Fills the sudoku display with the given sudoku
     * @param s The sudoku transport object to display
     */
    public void showSudokuValues(SudokuTransport s) {
        for (int x = 0; x < root * root; x++) {
            for (int y = 0; y < root * root; y++) {
                if (!cells[x][y].setValue(s.values[x][y]))
                    return;
            }
        }
        this.repaint();
        myWindow.repaint();
    }
    
    /**
     * Draws the backgrounds of the cells
     * @param type The type of the sudoku to display
     */
    public void drawBackgrounds(GameType type) {
        int size = root * root;
        for(SudokuCell[] row : cells)
            for(SudokuCell c : row)
                c.setBackground(null);
        switch (type) {
            case X:
                for (int i = 0; i < root * root; i++) {
                    cells[i][i].setBackground(Color.LIGHT_GRAY);
                    cells[i][root * root - 1 - i].setBackground(Color.LIGHT_GRAY);
                }
                break;
                
            case Hyper:
                for (int x = 1; x < size; x += root + 1) {
                    for (int y = 1; y < size; y += root + 1) {
                        for (int i = 0; i < root; i++) {
                            for (int j = 0; j < root; j++) {
                                cells[x + i][y + j].setBackground(Color.LIGHT_GRAY);
                            }
                        }
                    }
                }
                break;
                
            case Color:
                List<Color> colors = new ArrayList<>();
                for (int i = 0; i < root; i++)
                    for (int j = 0; j < root; j++)
                        colors.add(new Color(i * 128 / root + 127, j * 128 / root + 127, (Math.abs(i - j)) * 128 / root + 127));
                Random rnd = new Random();
                for (int i = 0; i < 150; i++) {
                    int a = rnd.nextInt(size);
                    int b = rnd.nextInt(size);
                    Color tmp = colors.get(a);
                    colors.set(a, colors.get(b));
                    colors.set(b, tmp);
                }
                for (int i = 0; i < size; i++) {
                    int xoff = i % root;
                    int yoff = i / root;
                    for (int jx = 0; jx < size; jx += root) {
                        for (int jy = 0; jy < size; jy += root) {
                            cells[xoff + jx][yoff + jy].setBackground(colors.get(i));
                        }
                    }
                }
                break;
        }
    }
    
    /**
     * Go one field right or jump to the beginning of the next row
     * @param x The x coordinate of the current cell
     * @param y The y coordinate of the current cell
     */
    public void goRight(int x, int y) {
        y++;
        if (y == root * root) {
            y = 0;
            x++;
            if (x == root * root)
                x = 0;
        }
        cells[x][y].requestFocusInWindow();
    }
    
    /**
     * Go one field left or jump to the end of the previous row
     * @param x The x coordinate of the current cell
     * @param y The y coordinate of the current cell
     */
    public void goLeft(int x, int y) {
        y--;
        if (y == -1) {
            y = root * root - 1;
            x--;
            if (x == -1)
                x = root * root - 1;
        }
        cells[x][y].requestFocusInWindow();
    }
    
    /**
     * Go one field up
     * @param x The x coordinate of the current cell
     * @param y The y coordinate of the current cell
     */
    public void goUp(int x, int y) {
        x--;
        if (x == -1) {
            x = root * root - 1;
        }
        cells[x][y].requestFocusInWindow();
    }
    
    /**
     * Go one field down
     * @param x The x coordinate of the current cell
     * @param y The y coordinate of the current cell
     */
    public void goDown(int x, int y) {
        x++;
        if (x == root * root) {
            x = 0;
        }
        cells[x][y].requestFocusInWindow();
    }
}
